/**
 * YARV: Yet Another Ruby VM.
 *
 *
 * $Id: yarvcore.c 275 2005-10-05 12:25:40Z ko1 $
 * Create : K.S. 04/01/01 01:17:22
 *
 * Copyright (c) 2004 SASADA Koichi <ko1 at atdot.net>
 */

#include <ruby.h>
#include <node.h>

#include "yarv_version.h"
#include "yarvcore.h"
#include "yarv.h"

VALUE mYarvCore;
VALUE cYarvISeq;
VALUE cYarvVM;
VALUE cYarvThread;
VALUE mYarvInsns;
VALUE cYarvEnv;
VALUE cYarvProc;

VALUE symIFUNC;
VALUE symCFUNC;

ID idPLUS;
ID idMINUS;
ID idMULT;
ID idDIV;
ID idMOD;
ID idLT;
ID idLTLT;
ID idEq;
ID idEqq;
ID idBackquote;
ID idEqTilde;
ID idThrowState;
ID idThrowBackDFP;
ID idThrowObject;
ID idAREF;
ID idASET;
ID idIntern;
ID idMethodMissing;
ID idEach;
ID idLength;

unsigned long yarvGlobalStateVersion = 1;


/* from Ruby 1.9 eval.c */
#ifdef HAVE_STDARG_PROTOTYPES
#include <stdarg.h>
#define va_init_list(a,b) va_start(a,b)
#else
#include <varargs.h>
#define va_init_list(a,b) va_start(a)
#endif

#define MARK_FREE_DEBUG 0
#if MARK_FREE_DEBUG
static int g_indent = 0;
static void gc_debug_indent(){
  int i;
  for(i=0; i<g_indent; i++){
    printf(" ");
  }
}
void MARK_REPORT(char *msg, int st){
  if(st == 0){
    g_indent--;
  }
  gc_debug_indent();
  printf("mark: %s\n", msg);
  if(st){
    g_indent++;
  }
}
#define GC_INFO gc_debug_indent(); printf
#define FREE_REPORT(msg) printf("free: %s\n", msg)

#else
#define MARK_REPORT(msg, st)
#define FREE_REPORT(msg)
#define GC_INFO if(0)printf
#endif

#define MARK_UNLESS_NULL(ptr) if(ptr){rb_gc_mark(ptr);}
#define FREE_UNLESS_NULL(ptr) if(ptr){ruby_xfree(ptr);}


/************/
/* YARVCore */
/************/

/* temporalily */
void yarv_setup(void *p1, void *p2, void *p3, void *p4,
                void *p5, void *p6, void *p7, void *p8);

yarv_thread_t *yarvCurrentThread = 0;
yarv_vm_t *theYarvVM = 0;
static VALUE yarvVMArray = Qnil;

VALUE
th_invoke_yield(yarv_thread_t *th, int argc, VALUE *argv);

/* rb_yield_values */
VALUE
yarv_yield_values(int argc, VALUE *argv)
{
  yarv_thread_t *th = GET_THREAD();

  if(argc == 1 && CLASS_OF(argv[0]) == rb_cValues){
    argc = RARRAY(argv[0])->len;
    argv = RARRAY(argv[0])->ptr;
  }
  
  return th_invoke_yield(th, argc, argv);
}

VALUE thread_call0(VALUE self, VALUE klass, VALUE recv, VALUE id, ID oid,
                   int argc, VALUE *argv, NODE *body, int nosuper);

/* rb_call0 continued for yarv function */
VALUE
yarv_call0(VALUE klass, VALUE recv, VALUE id, ID oid,
           int argc, VALUE *argv, NODE *body, int nosuper)
{
  return th_call0(GET_THREAD(), klass, recv, id, oid, argc, argv, body, nosuper);
}


VALUE th_call_super(yarv_thread_t *th, int argc, const VALUE *argv);

VALUE
yarv_call_super(int argc, const VALUE *argv)
{
  return th_call_super(GET_THREAD(), argc, argv);
}

VALUE th_backtrace(yarv_thread_t *th, int level);

VALUE
yarv_backtrace(int level)
{
  return th_backtrace(GET_THREAD(), level);
}

VALUE
yarv_caller(VALUE self, VALUE level)
{
  if(!IS_YARV_WORKING()){
    return rb_funcall(Qnil, rb_intern("caller"), 1, level);
  }
  else{
    return yarv_backtrace(FIX2INT(level));
  }
}

RUBY_EXTERN int rb_thread_critical;
RUBY_EXTERN int ruby_in_eval;
RUBY_EXTERN int ruby_nerrs;
RUBY_EXTERN NODE* ruby_eval_tree;

static VALUE vm_eval(VALUE self, VALUE iseq);

VALUE
yarv_load(char *file)
{
  NODE *node;
  VALUE argv[5];
  volatile VALUE iseq;
  volatile int critical;

  critical = rb_thread_critical;
  rb_thread_critical = Qtrue;
  {
    ruby_in_eval++;
    rb_load_file(file);
    ruby_in_eval--;
    node = ruby_eval_tree;
  }
  rb_thread_critical = critical;

  if(ruby_nerrs > 0){
    return 0;
  }
  
  argv[0] = (VALUE)node;
  argv[1] = rb_str_new2("<top (required)>");
  argv[2] = rb_str_new2(file);
  argv[3] = Qfalse;
  argv[4] = ISEQ_TYPE_TOP;
  iseq = rb_class_new_instance(5, argv, cYarvISeq);

  vm_eval(GET_VM()->self, iseq);
  return 0;
}


VALUE *th_svar(yarv_thread_t *self, int cnt);

VALUE *
rb_svar(int cnt)
{
  return th_svar(GET_THREAD(), cnt);
}

VALUE
rb_backref_get()
{
  VALUE *var = rb_svar(1);
  if (var) {
    return *var;
  }
  return Qnil;
}

void
rb_backref_set(VALUE val)
{
  VALUE *var = rb_svar(1);
  *var = val;
}

VALUE
rb_lastline_get()
{
  VALUE *var = rb_svar(0);
  if (var) {
    return *var;
  }
  return Qnil;
}

void
rb_lastline_set(VALUE val)
{
  VALUE *var = rb_svar(0);
  *var = val;
}


static VALUE
compile_string(VALUE str, VALUE file, VALUE line)
{
  NODE *node;
  node = rb_compile_string(StringValueCStr(file), str, NUM2INT(line));

  if(ruby_nerrs > 0){
    ruby_nerrs = 0;
    rb_exc_raise(GET_THREAD()->errinfo);
  }

  return (VALUE)node;
}


static VALUE
yarvcore_eval_iseq(VALUE iseq)
{
  VALUE vm;
  VALUE ret;
  vm = GET_VM()->self;
  ret = vm_eval(vm, iseq);
  return ret;
}

VALUE
yarvcore_eval_parsed(VALUE node, VALUE file)
{
  VALUE argv[5];
  VALUE iseq;
  
  argv[0] = node;
  argv[2] = file;
  
  if(GET_THREAD()->base_block){
    /* eval */
    argv[1] = GET_THREAD()->base_block->iseq->name;
    argv[3] = GET_THREAD()->base_block->iseq->self;
    argv[4] = ISEQ_TYPE_EVAL;
  }
  else{
    /* top */
    argv[1] = rb_str_new2("<main>");
    argv[3] = Qfalse;
    argv[4] = ISEQ_TYPE_TOP;
  }
  iseq = rb_class_new_instance(5, argv, cYarvISeq);
  return yarvcore_eval_iseq(iseq);
}

VALUE
yarvcore_eval(VALUE self, VALUE str, VALUE file, VALUE line)
{
  VALUE node;
  node = compile_string(str, file, line);
  return yarvcore_eval_parsed(node, file);
}

static VALUE yarvcore_parse(VALUE self, VALUE str, VALUE file, VALUE line){
  VALUE node;
  VALUE argv[5];
  VALUE iseq;
  
  node = compile_string(str, file, line);
  
  argv[0] = node;
  argv[1] = rb_str_new2("<main>");
  argv[2] = file;
  argv[3] = Qfalse;
  argv[4] = ISEQ_TYPE_TOP;
  iseq = rb_class_new_instance(5, argv, cYarvISeq);
  return iseq;
}


/***********************/
/* InstructionSequence */
/***********************/

static void
compile_data_free(struct iseq_compile_data *compile_data)
{
  if(compile_data){
    struct iseq_compile_data_storage *cur, *next;
    cur = compile_data->storage_head;
    while(cur){
      next = cur->next;
      ruby_xfree(cur);
      cur = next;
    }
    ruby_xfree(compile_data);
  }
}

static void
iseq_free(void *ptr)
{
  yarv_iseq_t *iseq;
  FREE_REPORT("-> iseq");
  
  if(ptr){
    iseq = ptr;
    GC_INFO("%s\n", RSTRING(iseq->name)->ptr);

    if(iseq->iseq != iseq->iseq_encoded){
      FREE_UNLESS_NULL(iseq->iseq_encoded);
    }
    FREE_UNLESS_NULL(iseq->iseq);

    FREE_UNLESS_NULL(iseq->insn_info_tbl);
    FREE_UNLESS_NULL(iseq->local_tbl);
    FREE_UNLESS_NULL(iseq->catch_table);
    FREE_UNLESS_NULL(iseq->arg_opt_tbl);
    compile_data_free(iseq->compile_data);
    ruby_xfree(ptr);
  }
  FREE_REPORT("<- iseq");
}

static void
iseq_mark(void *ptr)
{
  yarv_iseq_t *iseq;
  MARK_REPORT("-> iseq", 1);
  if(ptr){
    iseq = ptr;
    GC_INFO("%s\n", RSTRING(iseq->name)->ptr);
    MARK_UNLESS_NULL(iseq->iseq_mark_ary);
    MARK_UNLESS_NULL(iseq->name);
    MARK_UNLESS_NULL(iseq->file_name);
    MARK_UNLESS_NULL(iseq->klass_nest_stack);
    MARK_UNLESS_NULL(iseq->catch_table_ary);
    MARK_UNLESS_NULL((VALUE)iseq->node);
    
    if(iseq->compile_data != 0){
      MARK_UNLESS_NULL(iseq->compile_data->mark_ary);
      MARK_UNLESS_NULL(iseq->compile_data->err_info);
    }
  }
  MARK_REPORT("<- iseq", 0);
}

static VALUE
iseq_alloc(VALUE klass)
{
  VALUE volatile obj;
  yarv_iseq_t *iseq;

  obj = Data_Make_Struct(klass, yarv_iseq_t,
                         iseq_mark, iseq_free, iseq);
  return obj;
}

static VALUE
prepare_iseq_build(yarv_iseq_t *iseq,
                   VALUE name, VALUE file_name,
                   VALUE parent, VALUE type)
{
  iseq->name            = name;
  iseq->file_name       = file_name;
  iseq->iseq_mark_ary   = rb_ary_new();
  iseq->catch_table_ary = rb_ary_new();
  iseq->type            = type;
  iseq->arg_rest        = -1;
  iseq->arg_block       = -1;

  /* set class nest stack */
  if(parent){
    if(type != ISEQ_TYPE_CLASS){
      yarv_iseq_t *piseq;
      GetISeqVal(parent, piseq);
      iseq->klass_nest_stack = piseq->klass_nest_stack;
    }
  }
  else if(type == ISEQ_TYPE_TOP){
    iseq->klass_nest_stack = rb_ary_new();
    rb_ary_push(iseq->klass_nest_stack, rb_cObject);
  }
  else if(type == ISEQ_TYPE_METHOD){
    iseq->klass_nest_stack = rb_ary_new();
  }

  iseq->compile_data    = ALLOC(struct iseq_compile_data);
  MEMZERO(iseq->compile_data, struct iseq_compile_data, 1);
  iseq->compile_data->mark_ary = rb_ary_new();

  iseq->compile_data->storage_head = iseq->compile_data->storage_current =
    (struct iseq_compile_data_storage *)
      ALLOC_N(char, INITIAL_ISEQ_COMPILE_DATA_STORAGE_BUFF_SIZE +
              sizeof(struct iseq_compile_data_storage));

  iseq->compile_data->storage_head->pos  = 0;
  iseq->compile_data->storage_head->next = 0;
  iseq->compile_data->storage_head->size = INITIAL_ISEQ_COMPILE_DATA_STORAGE_BUFF_SIZE;
  iseq->compile_data->storage_head->buff = (char *)(&iseq->compile_data->storage_head->buff + 1);

  if(parent && CLASS_OF(parent) == cYarvISeq){
    yarv_iseq_t *piseq;
    GetISeqVal(parent, piseq);
    iseq->parent_iseq = piseq;

    while(piseq->parent_iseq){
      piseq = piseq->parent_iseq;
    }
    iseq->root_iseq = piseq;
  }
  return Qtrue;
}

static VALUE
cleanup_iseq_build(yarv_iseq_t *iseqobj)
{
  VALUE err;
  err = iseqobj->compile_data->err_info;
  compile_data_free(iseqobj->compile_data);
  iseqobj->compile_data = 0;

  if(ruby_nerrs > 0){
    ruby_nerrs = 0;
    rb_exc_raise(err);
  }
  return Qtrue;
}

static VALUE
iseq_init(VALUE self, VALUE node, VALUE name, VALUE file_name,
          VALUE parent, VALUE type)
{
  yarv_iseq_t *iseq;
  
  GetISeqVal(self, iseq);

  iseq->self = self;
  prepare_iseq_build(iseq, name, file_name, parent, type);
  iseq_compile(self, node);
  cleanup_iseq_build(iseq);
  return self;
}

VALUE iseq_assemble_setup(VALUE self, VALUE args, VALUE locals, VALUE ary);

static VALUE
iseq_assemble(VALUE self, VALUE args, VALUE locals, VALUE insn_ary)
{
  return iseq_assemble_setup(self, args, locals, insn_ary);
}

VALUE
iseq_inspect(VALUE self)
{
  char   buff[0x100];
  yarv_iseq_t *iseqobj;
  
  GetISeqVal(self, iseqobj);
  snprintf(buff, sizeof(buff), "<ISeq:%s@%s>",
           RSTRING(iseqobj->name)->ptr,
           RSTRING(iseqobj->file_name)->ptr);

  return rb_str_new2(buff);
}

/******/
/* VM */
/******/

void native_thread_cleanup(void *);

static void vm_free(void *ptr){
  FREE_REPORT("-> vm");
  if(ptr){
    yarv_vm_t *vmobj = ptr;
    
    native_thread_cleanup(ptr);

    st_free_table(vmobj->living_threads);
    ruby_xfree(ptr);
  }
  FREE_REPORT("<- vm");
}

static int
vm_mark_each_thread_func(st_data_t key, st_data_t value, st_data_t dummy)
{
  VALUE thval = (VALUE)key;
  rb_gc_mark(thval);
  return ST_CONTINUE;
}

static void vm_mark(void *ptr){
  MARK_REPORT("-> vm", 1);
  if(ptr){
    yarv_vm_t *vm = ptr;
    st_foreach(vm->living_threads, vm_mark_each_thread_func, 0);
  }
  MARK_REPORT("<- vm", 0);
}

static VALUE vm_alloc(VALUE klass){
  VALUE volatile obj;
  yarv_vm_t *vm;
  obj = Data_Make_Struct(klass, yarv_vm_t, vm_mark, vm_free, vm);
  return obj;
}

static VALUE vm_init(VALUE self, VALUE node){
  /* allocate vm stack */
  yarv_vm_t *vm;
  GetVMVal(self, vm);
  vm->self = self;
  return self;
}


static VALUE thread_eval(VALUE self, VALUE iseq);

static VALUE vm_eval(VALUE self, VALUE iseq){
  yarv_vm_t *vm;
  VALUE val;
  GetVMVal(self, vm);
  val = thread_eval(vm->main_thread, iseq);
  return val;
}


/**********/
/* Thread */
/**********/

static void thread_free(void *ptr){
  yarv_thread_t *th;

  FREE_REPORT("-> thread");
  if(ptr){
    th = ptr;
    FREE_UNLESS_NULL(th->stack);
    FREE_UNLESS_NULL(th->top_local_tbl);
    if(th->local_storage){
      st_free_table(th->local_storage);
    }
    ruby_xfree(ptr);
  }
  FREE_REPORT("<- thread");
}

void yarv_machine_stack_mark(yarv_thread_t *th);

static void thread_mark(void *ptr){
  yarv_thread_t *th;
  MARK_REPORT("-> thread", 1);
  if(ptr){
    th = ptr;
    if(th->stack){
      if(0){
        rb_gc_mark_locations(th->stack, th->cfp->sp);
      }
      else{
        VALUE *p = th->stack;
        VALUE *sp = th->cfp->sp;
        while(p < sp){
          rb_gc_mark(*p++);
        }
      }
    }
    
    /* mark ruby objects */
    MARK_UNLESS_NULL(th->klass_nest_stack);

    MARK_UNLESS_NULL(th->first_proc);
    MARK_UNLESS_NULL(th->first_args);

    MARK_UNLESS_NULL(th->wait_thread_value);
    MARK_UNLESS_NULL(th->value);
    MARK_UNLESS_NULL(th->errinfo);
    
    rb_mark_tbl(th->local_storage);

    if(GET_THREAD() != th &&
       th->machine_stack_start &&
       th->machine_stack_end){
      yarv_machine_stack_mark(th);
    }
  }
  
  MARK_UNLESS_NULL(th->stat_insn_usage);
  MARK_REPORT("<- thread", 0);
}

static VALUE thread_alloc(VALUE klass){
  VALUE volatile obj;
  yarv_thread_t *th;

  obj = Data_Make_Struct(klass, yarv_thread_t,
                         thread_mark, thread_free, th);
  return obj;
}

static void
th_init2(yarv_thread_t *th){
  MEMZERO(th, yarv_thread_t, 1);

  /* allocate thread stack */
  th->stack      = ALLOC_N(VALUE, YARV_THREAD_STACK_SIZE);
  th->stack_size = YARV_THREAD_STACK_SIZE;
  th->cfp        = (void *)(th->stack + th->stack_size);
  th->cfp--;
  
  th->cfp->pc  = 0;
  th->cfp->sp  = th->stack;
  th->cfp->bp  = 0;
  th->cfp->lfp = th->stack;
  th->cfp->dfp = th->stack;
  th->cfp->self = Qnil;
  th->cfp->magic = 0;
  
  th->status = THREAD_RUNNABLE;
  th->errinfo = Qnil;
}
static void
th_init(yarv_thread_t *th)
{
  th_init2(th);
  th->klass_nest_stack = rb_ary_push(rb_ary_new(), rb_cObject);
}

static VALUE
thread_init(VALUE self)
{
  yarv_thread_t *th;
  yarv_vm_t *vm = GET_VM();
  GetThreadVal(self, th);

  th_init(th);
  th->self = self;

  th->vm       = vm;
  return self;
}

VALUE th_eval_body(yarv_thread_t *th);
VALUE th_set_top_stack(yarv_thread_t *, VALUE iseq);
VALUE th_set_eval_stack(yarv_thread_t *, VALUE iseq);

static VALUE thread_eval(VALUE self, VALUE iseq){
  VALUE val;
  yarv_thread_t *th;
  GetThreadVal(self, th);

  if(th->base_block){
    th_set_eval_stack(th, iseq);
    th->base_block = 0;
  }
  else{
    th_set_top_stack(th, iseq);
  }
  val = th_eval_body(th);
  
  return val;
}


/***************/
/* YarvEnv     */
/***************/

static void
env_free(void *ptr)
{
  yarv_env_t *env;
  FREE_REPORT("-> env");
  if(ptr){
    env = ptr;
    FREE_UNLESS_NULL(env->env);
    ruby_xfree(ptr);
  }
  FREE_REPORT("<- env");
}

static void
env_mark(void *ptr)
{
  yarv_env_t *env;
  MARK_REPORT("-> env", 1);
  if(ptr){
    env = ptr;
    if(env->env){
      /* TODO: should mark more restricted range */
      rb_gc_mark_locations(env->env, env->env + env->env_size);
    }
    MARK_UNLESS_NULL(env->prev_envval);
    MARK_UNLESS_NULL(env->block.iseq->self);
  }
  MARK_REPORT("<- env", 0);
}

static VALUE
env_alloc(VALUE klass)
{
  VALUE obj;
  yarv_env_t *env;
  obj = Data_Make_Struct(klass, yarv_env_t,
                         env_mark, env_free, env);
  return obj;
}


/***************/
/* YarvProc    */
/***************/

static void
proc_free(void *ptr)
{
  FREE_REPORT("-> proc");
  if(ptr){
    ruby_xfree(ptr);
  }
  FREE_REPORT("<- proc");
}

static void
proc_mark(void *ptr)
{
  yarv_proc_t *proc;
  MARK_REPORT("-> proc", 1);
  if(ptr){
    proc = ptr;
    MARK_UNLESS_NULL(proc->envval);
    MARK_UNLESS_NULL(proc->blockprocval);
  }
  MARK_REPORT("<- proc", 0);
}

static VALUE
proc_alloc(VALUE klass)
{
  VALUE obj;
  yarv_proc_t *proc;
  obj = Data_Make_Struct(klass, yarv_proc_t,
                         proc_mark, proc_free, proc);
  return obj;
}

VALUE th_invoke_proc(yarv_thread_t *th, yarv_proc_t *proc, int argc, VALUE *argv);

static VALUE
proc_call(int argc, VALUE *argv, VALUE procval)
{
  yarv_proc_t *proc;
  GetProcVal(procval, proc);
  return th_invoke_proc(GET_THREAD(), proc, argc, argv);
}


/********************************************************************/

static VALUE yarv_once(){
  return rb_yield(Qnil);
}
static VALUE cfunc(){
  rb_funcall(Qnil, rb_intern("rfunc"), 0, 0);
  rb_funcall(Qnil, rb_intern("rfunc"), 0, 0);
  return Qnil;
}

VALUE yarv_Array_each();
VALUE yarv_Integer_times();
// VALUE yarv_Hash_each();
VALUE insns_name_array();
VALUE Init_yarvthread();
extern VALUE *rb_gc_stack_start;

char yarv_version[0x20];
char *yarv_options = ""
#if   OPT_DIRECT_THREADED_CODE
    "[direct threaded code] "
#elif OPT_TOKEN_THREADED_CODE
    "[token threaded code] "
#elif OPT_CALL_THREADED_CODE
    "[call threaded code] "
#endif

#if FAKE_INLINE_METHOD_CACHE
    "[fake inline method cache] "
#endif

#if OPT_BASIC_OPERATIONS
    "[optimize basic operation] "
#endif

#if OPT_STACK_CACHING
    "[stack caching] "
#endif

#if OPT_OPERANDS_UNIFICATION
    "[operands unification] "
#endif

#if OPT_INSTRUCTIONS_UNIFICATION
    "[instructions unification] "
#endif

#if OPT_INLINE_METHOD_CACHE
    "[inline method cache] "
#endif
      ;


void Init_yarvcore(){

#include "rev.inc"

  snprintf(yarv_version, 0x20, "YARVCore %d.%d.%d", MAJOR_VER, MINOR_VER, DEVEL_VER);
  
  /* declare YARVCore module */
  mYarvCore = rb_define_module("YARVCore");
  rb_define_const(mYarvCore, "VERSION", rb_str_new2(yarv_version));
  rb_define_const(mYarvCore, "MAJOR"  , INT2FIX(MAJOR_VER));
  rb_define_const(mYarvCore, "MINOR"  , INT2FIX(MINOR_VER));
  rb_define_const(mYarvCore, "REV"    , rb_str_new2(rev));
  rb_define_const(mYarvCore, "DATE"   , rb_str_new2(date));
  rb_define_const(mYarvCore, "OPTS"   , rb_str_new2(yarv_options));

  /* YARVCore::USAGE_ANALISYS_* */
  rb_define_const(mYarvCore, "USAGE_ANALISYS_INSN", rb_hash_new());
  rb_define_const(mYarvCore, "USAGE_ANALISYS_REGS", rb_hash_new());
  rb_define_const(mYarvCore, "USAGE_ANALISYS_INSN_BIGRAM", rb_hash_new());
  
  /* YARVCore::InsnNameArray */
  rb_define_const(mYarvCore, "InsnNameArray", insns_name_array());
  
  rb_define_singleton_method(mYarvCore, "eval" , yarvcore_eval , 3);
  rb_define_singleton_method(mYarvCore, "parse", yarvcore_parse, 3);
  rb_define_singleton_method(mYarvCore, "eval_iseq", yarvcore_eval_iseq, 1);
  
  /* declare YARVCore::InstructionSequence */
  cYarvISeq = rb_define_class_under(mYarvCore, "InstructionSequence", rb_cObject);
  rb_define_alloc_func(cYarvISeq, iseq_alloc);
  rb_define_method(cYarvISeq, "initialize", iseq_init, 5);
  rb_define_method(cYarvISeq, "inspect",    iseq_inspect, 0);
  rb_define_method(cYarvISeq, "disasm",     iseq_disasm, 0);
  rb_define_method(cYarvISeq, "assemble",   iseq_assemble, 3);
  // rb_define_singleton_method(cYarvISeq, "new_from_insn_ary", iseq_init_from_insn_ary, 1);
  
  /* declare YARVCore::VM */
  cYarvVM = rb_define_class_under(mYarvCore, "VM", rb_cObject);
  rb_define_alloc_func(cYarvVM, vm_alloc);
  rb_define_method(cYarvVM, "initialize", vm_init, 0);
  rb_define_method(cYarvVM, "eval",       vm_eval, 1);

  /* declare YARVCore::VM::Thread */
  cYarvThread = rb_define_class_under(cYarvVM, "Thread", rb_cObject);
  rb_define_global_const("Thread", cYarvThread);
  rb_define_alloc_func(cYarvThread, thread_alloc);
  rb_define_method(cYarvThread, "initialize", thread_init, 0);
  rb_define_method(cYarvThread, "eval",       thread_eval, 1);
  
  /* declare YARVCore::VM::Env */
  cYarvEnv = rb_define_class_under(cYarvVM, "Env", rb_cObject);
  rb_define_alloc_func(cYarvEnv, env_alloc);

  /* declare YARVCore::VM::Proc */
  cYarvProc = rb_define_class_under(cYarvVM, "Proc", rb_cObject);
  rb_define_alloc_func(cYarvProc, proc_alloc);
  rb_define_method(cYarvProc, "call", proc_call, -1);
  
  /* misc */
  
  
  /* YARV test functions */
  
  rb_define_global_function("once", yarv_once, 0);
  rb_define_global_function("cfunc", cfunc, 0);
  rb_define_global_function("yarv_caller", yarv_caller, 1);
  
  /* Integer#times */
  rb_define_method(rb_cInteger, "times", yarv_Integer_times, 0);
  
  /* Array#each */
  // rb_define_method(rb_cArray, "each", yarv_Array_each, 0);
  /* Hash#each */
  // rb_define_method(rb_cHash,"each", yarv_Hash_each, 0);

#if YARVEXT
  yarv_setup(yarv_yield_values, yarv_call0, yarv_block_given_p, yarv_load,
             yarv_svar, yarv_iterate, yarv_call_super, yarv_backtrace);
#endif
  
  symIFUNC = ID2SYM(rb_intern("<IFUNC>"));
  symCFUNC = ID2SYM(rb_intern("<CFUNC>"));

  /* for optimize */
  idPLUS = rb_intern("+");
  idMINUS= rb_intern("-");
  idMULT = rb_intern("*");
  idDIV  = rb_intern("/");
  idMOD  = rb_intern("%");
  idLT   = rb_intern("<");
  idLTLT = rb_intern("<<");
  idEq   = rb_intern("==");
  idEqq  = rb_intern("===");
  idBackquote = rb_intern("`");
  idEqTilde   = rb_intern("=~");
  
  idAREF = rb_intern("[]");
  idASET = rb_intern("[]=");

  idEach   = rb_intern("each");
  idLength = rb_intern("length");
  idIntern = rb_intern("intern");
  idMethodMissing = rb_intern("method_missing");

  idThrowState     = rb_intern("#__ThrowState__");
  idThrowBackDFP   = rb_intern("#__ThrowBackDFP__");
  idThrowObject    = rb_intern("#__ThrowObject__");
  
#if TEST_AOT_COMPILE
  Init_compiled();
#endif

  // make vm
  {
    /* create vm object */
    VALUE vmval = rb_class_new_instance(0, 0, cYarvVM);
    yarv_vm_t *vm;
    yarv_thread_t *th;
    GetVMVal(vmval, vm);
    theYarvVM = vm;

    yarvVMArray = rb_ary_new();
    rb_global_variable(&yarvVMArray);
    rb_ary_push(yarvVMArray, vm->self);

    /* create main thread */
    vm->main_thread = rb_class_new_instance(0, 0, cYarvThread);
    GetThreadVal(vm->main_thread, th);
    
    thread_free(GET_THREAD());
    yarv_set_current_running_thread(th);

    th->machine_stack_start = rb_gc_stack_start;
    Init_yarvthread();
    
    vm->living_threads = st_init_numtable();
    st_insert(vm->living_threads, th->self, (st_data_t)th->thread_id);
    /*
     * set toplevel binding
     */
  }
}

static void
test()
{
  int i; int *p;
  printf("!test!\n");
  for(i=0; i<1000000; i++){
    p = ALLOC(int);
  }
}

void
Init_yarv()
{
  /* initialize main thread */
  yarv_thread_t *th = ALLOC(yarv_thread_t);
  th_init2(th);
  yarv_set_current_running_thread(th);
}

